//
/*-
 * ---license-start
 * eu-digital-green-certificates / dgca-wallet-app-ios
 * ---
 * Copyright (C) 2021 T-Systems International GmbH and all other contributors
 * ---
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ---license-end
 */
//  
//  GatewayConnection.swift
//  DGCAWallet
//  
//  Created by Yannick Spreen on 5/3/21.
//  

import Foundation
import Alamofire
import SwiftDGC
import SwiftyJSON
import UIKit
import CertLogic
import CryptoKit

struct GatewayConnection: ContextConnection {
  public static func claim(cert: HCert, with tan: String?, completion: ((Bool, String?) -> Void)?) {
    guard var tan = tan, !tan.isEmpty else {
      return
    }
    // Replace dashes, spaces, etc. and turn into uppercase.
    let set = CharacterSet(charactersIn: "0123456789").union(.uppercaseLetters)
    tan = tan.uppercased().components(separatedBy: set.inverted).joined()

    let tanHash = SHA256.stringDigest(input: Data(tan.data(using: .utf8) ?? .init()))
    let certHash = cert.certHash
    let pubKey = (X509.derPubKey(for: cert.keyPair) ?? Data()).base64EncodedString()

    let toBeSigned = tanHash + certHash + pubKey
    let toBeSignedData = Data(toBeSigned.data(using: .utf8) ?? .init())
    Enclave.sign(data: toBeSignedData, with: cert.keyPair, using: .ecdsaSignatureMessageX962SHA256) { sign, err in
      guard let sign = sign, err == nil else {
        return
      }

      let keyParam: [String: Any] = [ "type": "EC", "value": pubKey ]
      let param: [String: Any] = [
        "DGCI": cert.uvci,
        "TANHash": tanHash,
        "certhash": certHash,
        "publicKey": keyParam,
        "signature": sign.base64EncodedString(),
        "sigAlg": "SHA256withECDSA"
      ]
      request(
        ["endpoints", "claim"],
        method: .post,
        parameters: param,
        encoding: JSONEncoding.default
      ).response {
        guard
          case .success(_) = $0.result,
          let status = $0.response?.statusCode,
          status / 100 == 2
        else {
          completion?(false, nil)
          return
        }

        let response = String(data: $0.data ?? .init(), encoding: .utf8)
        let json = JSON(parseJSON: response ?? "")
        let newTAN = json["tan"].string
        completion?(true, newTAN)
      }
    }
  }

  public static func fetchContext() {
    request(
      ["context"]
    ).response {
      guard
        let data = $0.data,
        let string = String(data: data, encoding: .utf8)
      else {
        return
      }
      let json = JSON(parseJSONC: string)
      LocalData.sharedInstance.config.merge(other: json)
      LocalData.sharedInstance.save()
      if LocalData.sharedInstance.versionedConfig["outdated"].bool == true {
        (
          UIApplication.shared.windows[0].rootViewController as? UINavigationController
        )?.popToRootViewController(animated: false)
      }
    }
  }
  static var config: JSON {
    LocalData.sharedInstance.versionedConfig
  }
}

// MARK: Country, Rules, Valuesets extension

extension GatewayConnection {
  // Country list
  public static func getListOfCountry(completion: (([CountryModel]) -> Void)?) {
    request(["endpoints", "countryList"], method: .get).response {
      guard
        case let .success(result) = $0.result,
        let response = result,
        let responseStr = String(data: response, encoding: .utf8),
        let json = JSON(parseJSON: responseStr).array
      else {
        return
      }
      let codes = json.compactMap { $0.string }
      var countryList: [CountryModel] = []
      codes.forEach { code in
        countryList.append(CountryModel(code: code))
      }
      completion?(countryList)
    }
  }
  static func countryList(completion: (([CountryModel]) -> Void)? = nil) {
    CountryDataStorage.initialize {
      if CountryDataStorage.sharedInstance.countryCodes.count > 0 {
        completion?(CountryDataStorage.sharedInstance.countryCodes.sorted(by: { countryOne, countryTwo in
          return countryOne.name < countryTwo.name
        }))
      }
      getListOfCountry { countryList in
        CountryDataStorage.sharedInstance.countryCodes.removeAll()
        countryList.forEach { country in
          CountryDataStorage.sharedInstance.add(country: country)
        }
        CountryDataStorage.sharedInstance.lastFetch = Date()
        CountryDataStorage.sharedInstance.save()
        completion?(CountryDataStorage.sharedInstance.countryCodes.sorted(by: { countryOne, countryTwo in
          return countryOne.name < countryTwo.name
        }))
      }
    }
  }
  // Rules
  public static func getListOfRules(completion: (([CertLogic.Rule]) -> Void)?) {
    request(["endpoints", "rules"], method: .get).response {
      guard
        case let .success(result) = $0.result,
        let response = result,
        let responseStr = String(data: response, encoding: .utf8)
      else {
        return
      }
      let ruleHashes: [RuleHash] = CertLogicEngine.getItems(from: responseStr)
      // Remove old hashes
      RulesDataStorage.sharedInstance.rules = RulesDataStorage.sharedInstance.rules.filter { rule in
          return !ruleHashes.contains(where: { ruleHash in
              return ruleHash.hash == rule.hash
          })
      }
      // Downloading new hashes
      var rulesItems = [CertLogic.Rule]()
      let downloadingGroup = DispatchGroup()
      ruleHashes.forEach { ruleHash in
        downloadingGroup.enter()
        if !RulesDataStorage.sharedInstance.isRuleExistWithHash(hash: ruleHash.hash) {
          getRules(ruleHash: ruleHash) { rule in
            if let rule = rule {
              rulesItems.append(rule)
            }
            downloadingGroup.leave()
          }
        } else {
          downloadingGroup.leave()
        }
      }
      downloadingGroup.notify(queue: .main) {
        completion?(rulesItems)
        print("Finished all requests.")
      }
    }
  }
  public static func getRules(ruleHash: CertLogic.RuleHash, completion: ((CertLogic.Rule?) -> Void)?) {
    request(["endpoints", "rules"], externalLink: "/\(ruleHash.country)/\(ruleHash.hash)", method: .get).response {
      guard
        case let .success(result) = $0.result,
        let response = result,
        let responseStr = String(data: response, encoding: .utf8)
      else {
        completion?(nil)
        return
      }
      if let rule: Rule = CertLogicEngine.getItem(from: responseStr) {
        let downloadedRuleHash = SHA256.digest(input: response as NSData)
        if downloadedRuleHash.hexString == ruleHash.hash {
          rule.setHash(hash: ruleHash.hash)
          completion?(rule)
        } else {
          completion?(nil)
        }
        return
      }
      completion?(nil)
    }
  }
  static func rulesList(completion: (([CertLogic.Rule]) -> Void)? = nil) {
    RulesDataStorage.initialize {
        completion?(RulesDataStorage.sharedInstance.rules)
    }
  }
  
  static func loadRulesFromServer(completion: (([CertLogic.Rule]) -> Void)? = nil) {
    getListOfRules { rulesList in
      rulesList.forEach { rule in
        RulesDataStorage.sharedInstance.add(rule: rule)
      }
      RulesDataStorage.sharedInstance.lastFetch = Date()
      RulesDataStorage.sharedInstance.save()
      completion?(RulesDataStorage.sharedInstance.rules)
    }
  }
  
  // ValueSets
  public static func getListOfValueSets(completion: (([CertLogic.ValueSet]) -> Void)?) {
    request(["endpoints", "valuesets"], method: .get).response {
      guard
        case let .success(result) = $0.result,
        let response = result,
        let responseStr = String(data: response, encoding: .utf8)
      else {
        return
      }
      let valueSetsHashes: [ValueSetHash] = CertLogicEngine.getItems(from: responseStr)
      // Remove old hashes
      ValueSetsDataStorage.sharedInstance.valueSets = ValueSetsDataStorage.sharedInstance.valueSets.filter { valueSet in
          return !valueSetsHashes.contains(where: { valueSetHashe in
              return valueSetHashe.hash == valueSet.hash
          })
      }
      // Downloading new hashes
      var valueSetsItems = [CertLogic.ValueSet]()
      let downloadingGroup = DispatchGroup()
      valueSetsHashes.forEach { valueSetHash in
        downloadingGroup.enter()
        if !ValueSetsDataStorage.sharedInstance.isValueSetExistWithHash(hash: valueSetHash.hash) {
          getValueSets(valueSetHash: valueSetHash) { valueSet in
            if let valueSet = valueSet {
              valueSetsItems.append(valueSet)
            }
            downloadingGroup.leave()
          }
        } else {
          downloadingGroup.leave()
        }
      }
      downloadingGroup.notify(queue: .main) {
        completion?(valueSetsItems)
        print("Finished all requests.")
      }
    }
  }
  public static func getValueSets(valueSetHash: CertLogic.ValueSetHash, completion: ((CertLogic.ValueSet?) -> Void)?) {
    request(["endpoints", "valuesets"], externalLink: "/\(valueSetHash.hash)", method: .get).response {
      guard
        case let .success(result) = $0.result,
        let response = result,
        let responseStr = String(data: response, encoding: .utf8)
      else {
        completion?(nil)
        return
      }
      if let valueSet: ValueSet = CertLogicEngine.getItem(from: responseStr) {
        let downloadedValueSetHash = SHA256.digest(input: response as NSData)
        if downloadedValueSetHash.hexString == valueSetHash.hash {
          valueSet.setHash(hash: valueSetHash.hash)
          completion?(valueSet)
        } else {
          completion?(nil)
        }
        return
      }
      completion?(nil)
    }
  }
  static func valueSetsList(completion: (([CertLogic.ValueSet]) -> Void)? = nil) {
    ValueSetsDataStorage.initialize {
        completion?(ValueSetsDataStorage.sharedInstance.valueSets)
    }
  }

  static func loadValueSetsFromServer(completion: (([CertLogic.ValueSet]) -> Void)? = nil){
    getListOfValueSets { valueSetsList in
      valueSetsList.forEach { valueSet in
        ValueSetsDataStorage.sharedInstance.add(valueSet: valueSet)
      }
      ValueSetsDataStorage.sharedInstance.lastFetch = Date()
      ValueSetsDataStorage.sharedInstance.save()
      completion?(ValueSetsDataStorage.sharedInstance.valueSets)
    }
  }
}

